@@ -19,18 +19,23 @@
 #include <sys/wait.h>
 
 #include "common.h"
-
+#include "io.h"
+#define NEED_CONF
 #include "config.h"
+#undef NEED_CONF
 
+void (*diag)(char *, ...);
+
+static const char *ident = "@(#) sacc(omys): " VERSION;
+
+static char intbuf[256]; /* 256B ought to be enough for any URI */
 static char *mainurl;
 static Item *mainentry;
 static int devnullfd;
 static int parent = 1;
 static int interactive;
 
-static void (*diag)(char *fmt, ...);
-
-void
+static void
 stddiag(char *fmt, ...)
 {
 	va_list arg;
@@ -103,7 +108,8 @@ mbsprint(const char *s, size_t len)
 {
 	wchar_t wc;
 	size_t col = 0, i, slen;
-	int rl, w;
+	const char *p;
+	int rl, pl, w;
 
 	if (!len)
 		return col;
@@ -112,27 +118,30 @@ mbsprint(const char *s, size_t len)
 	for (i = 0; i < slen; i += rl) {
 		rl = mbtowc(&wc, s + i, slen - i < 4 ? slen - i : 4);
 		if (rl == -1) {
-			mbtowc(NULL, NULL, 0); /* reset state */
-			fputs("\xef\xbf\xbd", stdout); /* replacement character */
-			col++;
-			rl = 1;
-			continue;
+			/* reset state */
+			mbtowc(NULL, NULL, 0);
+			p = "\xef\xbf\xbd"; /* replacement character */
+			pl = 3;
+			rl = w = 1;
+		} else {
+			if ((w = wcwidth(wc)) == -1)
+				continue;
+			pl = rl;
+			p = s + i;
 		}
-		if ((w = wcwidth(wc)) == -1)
-			continue;
 		if (col + w > len || (col + w == len && s[i + rl])) {
 			fputs("\xe2\x80\xa6", stdout); /* ellipsis */
 			col++;
 			break;
 		}
-		fwrite(s + i, 1, rl, stdout);
+		fwrite(p, 1, pl, stdout);
 		col += w;
 	}
 	return col;
 }
 
 static void *
-xreallocarray(void *m, const size_t n, const size_t s)
+xreallocarray(void *m, size_t n, size_t s)
 {
 	void *nm;
 
@@ -219,46 +228,83 @@ typedisplay(char t)
 {
 	switch (t) {
 	case '0':
-		return "Text+";
+		return typestr[TXT];
 	case '1':
-		return "Dir +";
+		return typestr[DIR];
 	case '2':
-		return "CSO |";
+		return typestr[CSO];
 	case '3':
-		return "Err |";
+		return typestr[ERR];
 	case '4':
-		return "Macf+";
+		return typestr[MAC];
 	case '5':
-		return "DOSf+";
+		return typestr[DOS];
 	case '6':
-		return "UUEf+";
+		return typestr[UUE];
 	case '7':
-		return "Find+";
+		return typestr[IND];
 	case '8':
-		return "Tlnt+";
+		return typestr[TLN];
 	case '9':
-		return "Binf+";
+		return typestr[BIN];
 	case '+':
-		return "Mirr+";
+		return typestr[MIR];
 	case 'T':
-		return "IBMt|";
+		return typestr[IBM];
 	case 'g':
-		return "GIF +";
+		return typestr[GIF];
 	case 'I':
-		return "Img +";
+		return typestr[IMG];
 	case 'h':
-		return "HTML+";
+		return typestr[URL];
 	case 'i':
-		return "    |";
+		return typestr[INF];
 	default:
 		/* "Characters '0' through 'Z' are reserved." (ASCII) */
 		if (t >= '0' && t <= 'Z')
-			return "!   |";
+			return typestr[BRK];
 		else
-			return "UNKN|";
+			return typestr[UNK];
 	}
 }
 
+int
+itemuri(Item *item, char *buf, size_t bsz)
+{
+	int n;
+
+	switch (item->type) {
+	case '8':
+		n = snprintf(buf, bsz, "telnet://%s@%s:%s",
+		             item->selector, item->host, item->port);
+		break;
+	case 'T':
+		n = snprintf(buf, bsz, "tn3270://%s@%s:%s",
+		             item->selector, item->host, item->port);
+		break;
+	case 'h':
+		n = snprintf(buf, bsz, "%s", item->selector +
+		             (strncmp(item->selector, "URL:", 4) ? 0 : 4));
+		break;
+	default:
+		n = snprintf(buf, bsz, "gopher://%s", item->host);
+
+		if (n < bsz-1 && strcmp(item->port, "70"))
+			n += snprintf(buf+n, bsz-n, ":%s", item->port);
+		if (n < bsz-1) {
+			n += snprintf(buf+n, bsz-n, "/%c%s",
+			              item->type, item->selector);
+		}
+		if (n < bsz-1 && item->type == '7' && item->tag) {
+			n += snprintf(buf+n, bsz-n, "%%09%s",
+			              item->tag + strlen(item->selector));
+		}
+		break;
+	}
+
+	return n;
+}
+
 static void
 printdir(Item *item)
 {
@@ -281,10 +327,17 @@ printdir(Item *item)
 static void
 displaytextitem(Item *item)
 {
+	struct sigaction sa;
 	FILE *pagerin;
 	int pid, wpid;
 
+	sigemptyset(&sa.sa_mask);
+	sa.sa_flags = SA_RESTART;
+	sa.sa_handler = SIG_DFL;
+	sigaction(SIGWINCH, &sa, NULL);
+
 	uicleanup();
+
 	switch (pid = fork()) {
 	case -1:
 		diag("Couldn't fork.");
@@ -300,6 +353,10 @@ displaytextitem(Item *item)
 			;
 	}
 	uisetup();
+
+	sa.sa_handler = uisigwinch;
+	sigaction(SIGWINCH, &sa, NULL);
+	uisigwinch(SIGWINCH); /* force redraw */
 }
 
 static char *
@@ -365,16 +422,13 @@ static Dir *
 molddiritem(char *raw)
 {
 	Item *item, *items = NULL;
-	char *s, *nl, *p;
+	char *nl, *p;
 	Dir *dir;
 	size_t i, n, nitems;
 
-	for (s = nl = raw, nitems = 0; p = strchr(nl, '\n'); ++nitems) {
-		s = nl;
-		nl = p+1;
-	}
-	if (!strcmp(s, ".\r\n") || !strcmp(s, ".\n"))
-		--nitems;
+	for (nl = raw, nitems = 0; p = strchr(nl, '\n'); nl = p+1)
+		++nitems;
+
 	if (!nitems) {
 		diag("Couldn't parse dir item");
 		return NULL;
@@ -405,7 +459,7 @@ molddiritem(char *raw)
 }
 
 static char *
-getrawitem(int sock)
+getrawitem(struct cnx *c)
 {
 	char *raw, *buf;
 	size_t bn, bs;
@@ -417,16 +471,28 @@ getrawitem(int sock)
 	do {
 		bs -= n;
 		buf += n;
+
+		if (buf - raw >= 5) {
+			if (strncmp(buf-5, "\r\n.\r\n", 5) == 0) {
+				buf[-3] = '\0';
+				break;
+			}
+		} else if (buf - raw == 3 && strncmp(buf-3, ".\r\n", 3)) {
+			buf[-3] = '\0';
+			break;
+		}
+
 		if (bs < 1) {
 			raw = xreallocarray(raw, ++bn, BUFSIZ);
 			buf = raw + (bn-1) * BUFSIZ;
 			bs = BUFSIZ;
 		}
-	} while ((n = read(sock, buf, bs)) > 0);
+
+	} while ((n = ioread(c, buf, bs)) > 0);
 
 	*buf = '\0';
 
-	if (n < 0) {
+	if (n == -1) {
 		diag("Can't read socket: %s", strerror(errno));
 		clear(&raw);
 	}
@@ -435,7 +501,7 @@ getrawitem(int sock)
 }
 
 static int
-sendselector(int sock, const char *selector)
+sendselector(struct cnx *c, const char *selector)
 {
 	char *msg, *p;
 	size_t ln;
@@ -445,10 +511,10 @@ sendselector(int sock, const char *selector)
 	msg = p = xmalloc(ln);
 	snprintf(msg, ln--, "%s\r\n", selector);
 
-	while ((n = write(sock, p, ln)) > 0) {
+	while ((n = iowrite(c, p, ln)) > 0) {
 		ln -= n;
 		p += n;
-	}
+	};
 
 	free(msg);
 	if (n == -1)
@@ -458,7 +524,7 @@ sendselector(int sock, const char *selector)
 }
 
 static int
-connectto(const char *host, const char *port)
+connectto(const char *host, const char *port, struct cnx *c)
 {
 	sigset_t set, oset;
 	static const struct addrinfo hints = {
@@ -466,8 +532,8 @@ connectto(const char *host, const char *port)
 	    .ai_socktype = SOCK_STREAM,
 	    .ai_protocol = IPPROTO_TCP,
 	};
-	struct addrinfo *addrs, *addr;
-	int r, sock = -1;
+	struct addrinfo *addrs, *ai;
+	int r, err;
 
 	sigemptyset(&set);
 	sigaddset(&set, SIGWINCH);
@@ -479,68 +545,68 @@ connectto(const char *host, const char *port)
 		goto err;
 	}
 
-	for (addr = addrs; addr; addr = addr->ai_next) {
-		if ((sock = socket(addr->ai_family, addr->ai_socktype,
-		                   addr->ai_protocol)) < 0)
-			continue;
-		if ((r = connect(sock, addr->ai_addr, addr->ai_addrlen)) < 0) {
-			close(sock);
-			continue;
-		}
-		break;
+	r = -1;
+	for (ai = addrs; ai && r == -1; ai = ai->ai_next) {
+		do {
+			if ((c->sock = socket(ai->ai_family, ai->ai_socktype,
+			                      ai->ai_protocol)) == -1) {
+				err = errno;
+				break;
+			}
+
+			if ((r = ioconnect(c, ai, host)) < 0) {
+				err = errno;
+				ioclose(c);
+			}
+		/* retry on cleartext */
+		} while (r == -2);
 	}
 
 	freeaddrinfo(addrs);
 
-	if (sock < 0) {
-		diag("Can't open socket: %s", strerror(errno));
-		goto err;
-	}
-	if (r < 0) {
-		diag("Can't connect to: %s:%s: %s",
-		     host, port, strerror(errno));
-		goto err;
-	}
-
-	sigprocmask(SIG_SETMASK, &oset, NULL);
-	return sock;
-
+	if (r == -1)
+		ioconnerr(c, host, port, err);
 err:
 	sigprocmask(SIG_SETMASK, &oset, NULL);
-	return -1;
+
+	return r;
 }
 
 static int
 download(Item *item, int dest)
 {
 	char buf[BUFSIZ];
+	struct cnx c;
 	ssize_t r, w;
-	int src;
 
-	if (!item->tag) {
-		if ((src = connectto(item->host, item->port)) < 0 ||
-		    sendselector(src, item->selector) < 0)
+	if (item->tag == NULL) {
+		if (connectto(item->host, item->port, &c) < 0 ||
+		    sendselector(&c, item->selector) == -1)
 			return 0;
-	} else if ((src = open(item->tag, O_RDONLY)) < 0) {
-		printf("Can't open source file %s: %s",
-		       item->tag, strerror(errno));
-		errno = 0;
-		return 0;
+	} else {
+		if ((c.sock = open(item->tag, O_RDONLY)) == -1) {
+			printf("Can't open source file %s: %s",
+			       item->tag, strerror(errno));
+			errno = 0;
+			return 0;
+		}
+		c.tls = NULL;
 	}
 
 	w = 0;
-	while ((r = read(src, buf, BUFSIZ)) > 0) {
+	while ((r = ioread(&c, buf, BUFSIZ)) > 0) {
 		while ((w = write(dest, buf, r)) > 0)
 			r -= w;
 	}
 
-	if (r < 0 || w < 0) {
+	if (r == -1 || w == -1) {
 		printf("Error downloading file %s: %s",
 		       item->selector, strerror(errno));
 		errno = 0;
 	}
 
-	close(src);
+	close(dest);
+	ioclose(&c);
 
 	return (r == 0 && w == 0);
 }
@@ -564,14 +630,14 @@ downloaditem(Item *item)
 		path = xstrdup(file);
 
 	if (tag = item->tag) {
-		if (access(tag, R_OK) < 0) {
+		if (access(tag, R_OK) == -1) {
 			clear(&item->tag);
 		} else if (!strcmp(tag, path)) {
 			goto cleanup;
 		}
 	}
 
-	if ((dest = open(path, O_WRONLY|O_CREAT|O_EXCL, mode)) < 0) {
+	if ((dest = open(path, O_WRONLY|O_CREAT|O_EXCL, mode)) == -1) {
 		diag("Can't open destination file %s: %s",
 		     path, strerror(errno));
 		errno = 0;
@@ -595,24 +661,41 @@ cleanup:
 static int
 fetchitem(Item *item)
 {
-	int sock;
+	struct cnx c;
+	char *raw;
 
-	if ((sock = connectto(item->host, item->port)) < 0 ||
-	    sendselector(sock, item->selector) < 0)
+	if (connectto(item->host, item->port, &c) < 0 ||
+	    sendselector(&c, item->selector) == -1)
 		return 0;
-	item->raw = getrawitem(sock);
-	close(sock);
 
-	if (item->raw && !*item->raw) {
+	raw = getrawitem(&c);
+	ioclose(&c);
+
+	if (raw == NULL || !*raw) {
 		diag("Empty response from server");
-		clear(&item->raw);
+		clear(&raw);
 	}
 
-	return (item->raw != NULL);
+	return ((item->raw = raw) != NULL);
 }
 
 static void
-plumb(char *url)
+pipeuri(char *cmd, char *msg, char *uri)
+{
+	FILE *sel;
+
+	if ((sel = popen(cmd, "w")) == NULL) {
+		diag("URI not %s\n", msg);
+		return;
+	}
+
+	fputs(uri, sel);
+	pclose(sel);
+	diag("%s \"%s\"", msg, uri);
+}
+
+static void
+execuri(char *cmd, char *msg, char *uri)
 {
 	switch (fork()) {
 	case -1:
@@ -622,11 +705,16 @@ plumb(char *url)
 		parent = 0;
 		dup2(devnullfd, 1);
 		dup2(devnullfd, 2);
-		if (execlp(plumber, plumber, url, NULL) < 0)
+		if (execlp(cmd, cmd, uri, NULL) == -1)
 			_exit(1);
+	default:
+		if (modalplumber) {
+			while (waitpid(-1, NULL, 0) != -1)
+				;
+		}
 	}
 
-	diag("Plumbed \"%s\"", url);
+	diag("%s \"%s\"", msg, uri);
 }
 
 static void
@@ -646,7 +734,7 @@ plumbitem(Item *item)
 	if (!path)
 		return;
 
-	if ((tag = item->tag) && access(tag, R_OK) < 0) {
+	if ((tag = item->tag) && access(tag, R_OK) == -1) {
 		clear(&item->tag);
 		tag = NULL;
 	}
@@ -656,14 +744,14 @@ plumbitem(Item *item)
 	if (!path[0]) {
 		clear(&path);
 		if (!tag) {
-			if (asprintf(&path, "%s/%s", tmpdir, file) < 0)
+			if (asprintf(&path, "%s/%s", tmpdir, file) == -1)
 				die("Can't generate tmpdir path: %s/%s: %s",
 				    tmpdir, file, strerror(errno));
 		}
 	}
 
 	if (path && (!tag || strcmp(tag, path))) {
-		if ((dest = open(path, O_WRONLY|O_CREAT|O_EXCL, mode)) < 0) {
+		if ((dest = open(path, O_WRONLY|O_CREAT|O_EXCL, mode)) == -1) {
 			diag("Can't open destination file %s: %s",
 			     path, strerror(errno));
 			errno = 0;
@@ -677,7 +765,7 @@ plumbitem(Item *item)
 		item->tag = path;
 
 	if (plumbitem)
-		plumb(item->tag);
+		execuri(plumber, "Plumbed", item->tag);
 
 	return;
 cleanup:
@@ -685,6 +773,13 @@ cleanup:
 	return;
 }
 
+void
+yankitem(Item *item)
+{
+	itemuri(item, intbuf, sizeof(intbuf));
+	pipeuri(yanker, "Yanked", intbuf);
+}
+
 static int
 dig(Item *entry, Item *item)
 {
@@ -700,7 +795,7 @@ dig(Item *entry, Item *item)
 	switch (t) {
 	case 'h': /* fallthrough */
 		if (!strncmp(item->selector, "URL:", 4)) {
-			plumb(item->selector+4);
+			execuri(plumber, "Plumbed", item->selector+4);
 			return 0;
 		}
 	case '0':
@@ -721,17 +816,17 @@ dig(Item *entry, Item *item)
 	case '8':
 		if (asprintf(&plumburi, "telnet://%s%s%s:%s",
 		             item->selector, item->selector ? "@" : "",
-		             item->host, item->port) < 0)
+		             item->host, item->port) == -1)
 			return 0;
-		plumb(plumburi);
+		execuri(plumber, "Plumbed", plumburi);
 		free(plumburi);
 		return 0;
 	case 'T':
 		if (asprintf(&plumburi, "tn3270://%s%s%s:%s",
 		             item->selector, item->selector ? "@" : "",
-		             item->host, item->port) < 0)
+		             item->host, item->port) == -1)
 			return 0;
-		plumb(plumburi);
+		execuri(plumburi, "Plumbed", plumburi);
 		free(plumburi);
 		return 0;
 	default:
@@ -885,11 +980,7 @@ moldentry(char *url)
 	char *p, *host = url, *port = "70", *gopherpath = "1";
 	int parsed, ipv6;
 
-	if (p = strstr(url, "://")) {
-		if (strncmp(url, "gopher", p - url))
-			die("Protocol not supported: %.*s", p - url, url);
-		host = p + 3;
-	}
+	host = ioparseurl(url);
 
 	if (*host == '[') {
 		ipv6 = 1;
@@ -957,7 +1048,7 @@ cleanup(void)
 		uicleanup();
 }
 
-void
+static void
 sighandler(int signo)
 {
 	exit(128 + signo);
@@ -973,12 +1064,12 @@ setup(void)
 	setenv("PAGER", "more", 0);
 	atexit(cleanup);
 	/* reopen stdin in case we're reading from a pipe */
-	if ((fd = open("/dev/tty", O_RDONLY)) < 0)
+	if ((fd = open("/dev/tty", O_RDONLY)) == -1)
 		die("open: /dev/tty: %s", strerror(errno));
-	if (dup2(fd, 0) < 0)
+	if (dup2(fd, 0) == -1)
 		die("dup2: /dev/tty, stdin: %s", strerror(errno));
 	close(fd);
-	if ((devnullfd = open("/dev/null", O_WRONLY)) < 0)
+	if ((devnullfd = open("/dev/null", O_WRONLY)) == -1)
 		die("open: /dev/null: %s", strerror(errno));
 
 	sigemptyset(&sa.sa_mask);
@@ -990,10 +1081,13 @@ setup(void)
 
 	if (!mkdtemp(tmpdir))
 		die("mkdir: %s: %s", tmpdir, strerror(errno));
-	if(interactive = isatty(1)) {
+	if (interactive = isatty(1)) {
 		uisetup();
+		diag = uistatus;
 		sa.sa_handler = uisigwinch;
 		sigaction(SIGWINCH, &sa, NULL);
+	} else {
+		diag = stddiag;
 	}
 }
 
@@ -1006,15 +1100,12 @@ main(int argc, char *argv[])
 	setup();
 
 	mainurl = xstrdup(argv[1]);
-
 	mainentry = moldentry(mainurl);
-	if (interactive) {
-		diag = uistatus;
+
+	if (interactive)
 		delve(mainentry);
-	} else {
-		diag = stddiag;
+	else
 		printout(mainentry);
-	}
 
 	exit(0);
 }
