@@ -7,16 +7,25 @@
 #include <unistd.h>
 #include <sys/types.h>
 
-#include "config.h"
 #include "common.h"
+#include "config.h"
 
 #define C(c) #c
 #define S(c) C(c)
 
+/* ncurses doesn't define those in term.h, where they're used */
+#ifndef OK
+#define OK (0)
+#endif
+#ifndef ERR
+#define ERR (-1)
+#endif
+
 static char bufout[256];
 static struct termios tsave;
 static struct termios tsacc;
 static Item *curentry;
+static int termset = ERR;
 
 void
 uisetup(void)
@@ -24,9 +33,13 @@ uisetup(void)
 	tcgetattr(0, &tsave);
 	tsacc = tsave;
 	tsacc.c_lflag &= ~(ECHO|ICANON);
+	tsacc.c_cc[VMIN] = 1;
+	tsacc.c_cc[VTIME] = 0;
 	tcsetattr(0, TCSANOW, &tsacc);
 
-	setupterm(NULL, 1, NULL);
+	if (termset != OK)
+		/* setupterm call exits on error */
+		termset = setupterm(NULL, 1, NULL);
 	putp(tparm(clear_screen, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 	putp(tparm(save_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 	putp(tparm(change_scroll_region, 0, lines-2, 0, 0, 0, 0, 0, 0, 0));
@@ -37,9 +50,13 @@ uisetup(void)
 void
 uicleanup(void)
 {
+	tcsetattr(0, TCSANOW, &tsave);
+
+	if (termset != OK)
+		return;
+
 	putp(tparm(change_scroll_region, 0, lines-1, 0, 0, 0, 0, 0, 0, 0));
 	putp(tparm(clear_screen, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	tcsetattr(0, TCSANOW, &tsave);
 	fflush(stdout);
 }
 
@@ -58,14 +75,13 @@ uiprompt(char *fmt, ...)
 	putp(tparm(enter_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 
 	va_start(ap, fmt);
-	if (vsnprintf(bufout, sizeof(bufout), fmt, ap) >= sizeof(bufout))
-		bufout[sizeof(bufout)-1] = '\0';
+	vsnprintf(bufout, sizeof(bufout), fmt, ap);
 	va_end(ap);
+
 	n = mbsprint(bufout, columns);
 
 	putp(tparm(exit_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	if (n < columns)
-		printf("%*s", columns - n, " ");
+	putp(tparm(clr_eol, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 
 	putp(tparm(cursor_address, lines-1, n, 0, 0, 0, 0, 0, 0, 0));
 
@@ -81,7 +97,7 @@ uiprompt(char *fmt, ...)
 	putp(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 	fflush(stdout);
 
-	if (r < 0) {
+	if (r == -1) {
 		clearerr(stdin);
 		clear(&input);
 	} else if (input[r - 1] == '\n') {
@@ -94,9 +110,9 @@ uiprompt(char *fmt, ...)
 static void
 printitem(Item *item)
 {
-	if (snprintf(bufout, sizeof(bufout), "%s %s", typedisplay(item->type),
-	    item->username) >= sizeof(bufout))
-		bufout[sizeof(bufout)-1] = '\0';
+	snprintf(bufout, sizeof(bufout), "%s %s",
+	         typedisplay(item->type), item->username);
+
 	mbsprint(bufout, columns);
 	putchar('\r');
 }
@@ -122,6 +138,8 @@ help(Item *entry)
 		       S(_key_searchprev) ": search string backward.\n"
 		       S(_key_cururi) ": print page URI.\n"
 		       S(_key_seluri) ": print item URI.\n"
+		       S(_key_yankcur) ": yank page URI to external program.\n"
+		       S(_key_yanksel) ": yank item URI to external program.\n"
 		       S(_key_help) ": show this help.\n"
 		       "^D, " S(_key_quit) ": exit sacc.\n"
 	};
@@ -147,16 +165,14 @@ uistatus(char *fmt, ...)
 	va_end(ap);
 
 	if (n < sizeof(bufout)-1) {
-		n += snprintf(bufout + n, sizeof(bufout) - n,
-		              " [Press a key to continue \xe2\x98\x83]");
+		snprintf(bufout+n, sizeof(bufout)-n,
+		         " [Press a key to continue \xe2\x98\x83]");
 	}
-	if (n >= sizeof(bufout))
-		bufout[sizeof(bufout)-1] = '\0';
 
-	n = mbsprint(bufout, columns);
+	mbsprint(bufout, columns);
+
 	putp(tparm(exit_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	if (n < columns)
-		printf("%*s", columns - n, " ");
+	putp(tparm(clr_eol, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 
 	putp(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 	fflush(stdout);
@@ -169,25 +185,25 @@ displaystatus(Item *item)
 {
 	Dir *dir = item->dat;
 	char *fmt;
-	size_t n, nitems = dir ? dir->nitems : 0;
+	size_t nitems = dir ? dir->nitems : 0;
 	unsigned long long printoff = dir ? dir->printoff : 0;
 
 	putp(tparm(save_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 
 	putp(tparm(cursor_address, lines-1, 0, 0, 0, 0, 0, 0, 0, 0));
 	putp(tparm(enter_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
+
 	fmt = (strcmp(item->port, "70") && strcmp(item->port, "gopher")) ?
 	      "%1$3lld%%| %2$s:%5$s/%3$c%4$s" : "%3lld%%| %s/%c%s";
-	if (snprintf(bufout, sizeof(bufout), fmt,
-	             (printoff + lines-1 >= nitems) ? 100 :
-	             (printoff + lines-1) * 100 / nitems,
-	             item->host, item->type, item->selector, item->port)
-	    >= sizeof(bufout))
-		bufout[sizeof(bufout)-1] = '\0';
-	n = mbsprint(bufout, columns);
+	snprintf(bufout, sizeof(bufout), fmt,
+	         (printoff + lines-1 >= nitems) ? 100 :
+	         (printoff + lines-1) * 100 / nitems,
+	         item->host, item->type, item->selector, item->port);
+
+	mbsprint(bufout, columns);
+
 	putp(tparm(exit_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	if (n < columns)
-		printf("%*s", columns - n, " ");
+	putp(tparm(clr_eol, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 
 	putp(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 	fflush(stdout);
@@ -196,8 +212,6 @@ displaystatus(Item *item)
 static void
 displayuri(Item *item)
 {
-	size_t n;
-
 	if (item->type == 0 || item->type == 'i')
 		return;
 
@@ -205,44 +219,13 @@ displayuri(Item *item)
 
 	putp(tparm(cursor_address, lines-1, 0, 0, 0, 0, 0, 0, 0, 0));
 	putp(tparm(enter_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	switch (item->type) {
-	case '8':
-		n = snprintf(bufout, sizeof(bufout), "telnet://%s@%s:%s",
-		             item->selector, item->host, item->port);
-		break;
-	case 'h':
-		n = snprintf(bufout, sizeof(bufout), "%s",
-		             item->selector);
-		break;
-	case 'T':
-		n = snprintf(bufout, sizeof(bufout), "tn3270://%s@%s:%s",
-		             item->selector, item->host, item->port);
-		break;
-	default:
-		n = snprintf(bufout, sizeof(bufout), "gopher://%s", item->host);
 
-		if (n < sizeof(bufout) && strcmp(item->port, "70")) {
-			n += snprintf(bufout+n, sizeof(bufout)-n, ":%s",
-			              item->port);
-		}
-		if (n < sizeof(bufout)) {
-			n += snprintf(bufout+n, sizeof(bufout)-n, "/%c%s",
-			              item->type, item->selector);
-		}
-		if (n < sizeof(bufout) && item->type == '7' && item->tag) {
-			n += snprintf(bufout+n, sizeof(bufout)-n, "%%09%s",
-			              item->tag + strlen(item->selector));
-		}
-		break;
-	}
+	itemuri(item, bufout, sizeof(bufout));
 
-	if (n >= sizeof(bufout))
-		bufout[sizeof(bufout)-1] = '\0';
+	mbsprint(bufout, columns);
 
-	n = mbsprint(bufout, columns);
 	putp(tparm(exit_standout_mode, 0, 0, 0, 0, 0, 0, 0, 0, 0));
-	if (n < columns)
-		printf("%*s", columns - n, " ");
+	putp(tparm(clr_eol, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 
 	putp(tparm(restore_cursor, 0, 0, 0, 0, 0, 0, 0, 0, 0));
 	fflush(stdout);
@@ -391,7 +374,7 @@ jumptoline(Item *entry, ssize_t line, int absolute)
 	return;
 }
 
-void
+static void
 searchinline(const char *searchstr, Item *entry, int pos)
 {
 	Dir *dir;
@@ -524,7 +507,6 @@ uiselectitem(Item *entry)
 			jumptoline(entry, 0, 0);
 			continue;
 		case _key_search:
-		search:
 			free(searchstr);
 			if (!((searchstr = uiprompt("Search for: ")) &&
 			    searchstr[0])) {
@@ -542,7 +524,6 @@ uiselectitem(Item *entry)
 		quit:
 			return NULL;
 		case _key_fetch:
-		fetch:
 			if (entry->raw)
 				continue;
 			return entry;
@@ -554,6 +535,14 @@ uiselectitem(Item *entry)
 			if (dir)
 				displayuri(&dir->items[dir->curline]);
 			continue;
+		case _key_yankcur:
+			if (dir)
+				yankitem(entry);
+			continue;
+		case _key_yanksel:
+			if (dir)
+				yankitem(&dir->items[dir->curline]);
+			continue;
 		case _key_help: /* FALLTHROUGH */
 			return help(entry);
 		default:
@@ -567,14 +556,16 @@ uisigwinch(int signal)
 {
 	Dir *dir;
 
-	setupterm(NULL, 1, NULL);
+	if (termset == OK)
+		del_curterm(cur_term);
+	termset = setupterm(NULL, 1, NULL);
 	putp(tparm(change_scroll_region, 0, lines-2, 0, 0, 0, 0, 0, 0, 0));
 
 	if (!curentry || !(dir = curentry->dat))
 		return;
 
 	if (dir->curline - dir->printoff > lines-2)
-		dir->curline = dir->printoff + lines-2;
+		dir->printoff = dir->curline - (lines-2);
 
 	uidisplay(curentry);
 }
